/*
 * (c) 2024 Raphael Mudge
 */
package crystalpalace.coff;

import crystalpalace.util.*;
import java.util.*;

/*
 * References:
 * https://courses.cs.washington.edu/courses/cse378/03wi/lectures/LinkerFiles/coff.pdf
 * https://learn.microsoft.com/en-us/windows/win32/debug/pe-format
 */
public class COFFObject {
	protected int         Machine;
	protected int         NumberOfSections;
	protected long        TimeDateStamp;
	protected long        PointerToSymbolTable;
	protected long        NumberOfSymbols;
	protected int         SizeOfOptionalHeader;
	protected int         Characteristics;
	protected List        sections = new LinkedList();
	protected List        symbols  = new LinkedList();
	protected StringTable strings;

	public COFFObject(ByteWalker walker) {
		Machine              = walker.readShort();

		/* sanity check the header */
		if (!CrystalUtils.toSet("arm64, x86, x64").contains(getMachine()))
			throw new RuntimeException("COFF starts with unrecognized Machine value " + CrystalUtils.toHex(Machine));

		NumberOfSections     = walker.readShort();
		TimeDateStamp        = walker.readInt();
		PointerToSymbolTable = walker.readInt();
		NumberOfSymbols      = walker.readInt();
		SizeOfOptionalHeader = walker.readShort();
		Characteristics      = walker.readShort();

		/* skip our optional header */
		walker.skip(SizeOfOptionalHeader);

		/* we now have our section table */
		for (int x = 0; x < NumberOfSections; x++) {
			sections.add(new Section(this, walker));
		}

		/* we now have our symbol table */
		walker.GoTo((int)PointerToSymbolTable);

		for (int x = 0; x < NumberOfSymbols; x++) {
			symbols.add(new Symbol(x, this, walker));
		}

		/* immediately following the symbol table, we have... the string table */
		strings = new StringTable(this, walker);

		walker.Return();

		/* if we had errors, then register our dissatisfaction now */
		if (!walker.isSane())
			throw new RuntimeException("Error(s) while parsing COFF. File is possibly malformed.");
	}

	public Symbol getSymbol(String name) {
		Iterator i = symbols.iterator();
		while (i.hasNext()) {
			Symbol next = (Symbol)i.next();
			if ( name.equals(next.getName()) )
				return next;
		}

		return null;
	}

	public Symbol getSymbol(int index) {
		Iterator i = symbols.iterator();
		while (i.hasNext()) {
			Symbol next = (Symbol)i.next();
			if (next.is(index))
				return next;
		}

		return null;
	}

	/* patch the  value of the specified symbol within our COFF with whatever we want */
	public void patch(String name, byte[] data) {
		Symbol temp = getSymbol(name);
		if (temp == null)
			throw new RuntimeException("No symbol '" + name + "'");

		Section sect = temp.getSection();

		/* check that our section is initialized, if it's not... throw a tantrum */
		if (sect.isUninitialized())
			throw new RuntimeException("Can't patch symbol " + name + " in uninitialized " + sect.getName() + " section");

		/* estimate the size of the symbol, this is to help sanity check the patch */
		long est = temp.estimateSize();

		/* this is likely a sane and accurate estimate, and maybe a pointer, so we want our patch to fill it
		   in completely */
		if (est == 4 || est == 8) {
			if (est != data.length)
				throw new RuntimeException("Symbol " + name + " (est.) size " + est + "b differs from patch " + data.length + "b size");
		}
		/* for other symbol sizes... we want to just validate that we're not exceeding the estimated size. I've noticed
		 * for oddball global vars (e.g., 40b or something)--the compiler may round up the space allotted for that data
		 * in whatever section. So this is just allowing for that, but checking what we can anyways */
		else {
			if (data.length > est)
				throw new RuntimeException("Symbol " + name + " (est.) size " + est + "b is LESS than patch " + data.length + "b size");
		}

		/* OK, now we can do the patch */
		sect.patch((int)temp.getValue(), data);
	}

	public String getStringAt(int offset) {
		return strings.getStringAt(offset);
	}

	public String getMachine() {
		switch (Machine) {
			case 0x8664:
				return "x64";
			case 0x14c:
				return "x86";
			case 0xaa64:
				return "arm64";
			default:
				return "Unknown " + CrystalUtils.toHex(Machine);
		}
	}

	public List getSymbolsForSection(int req) {
		List result = new LinkedList(symbols);
		Iterator i = result.iterator();
		while (i.hasNext()) {
			Symbol temp = (Symbol)i.next();
			if (temp.SectionNumber != req)
				i.remove();
		}

		Collections.sort(result);

		return result;
	}

	public Section getSection(int req) {
		Iterator i = sections.iterator();
		for (int x = 1; i.hasNext(); x++) {
			Section next = (Section)i.next();
			if (x == req)
				return next;
		}

		return null;
	}

	public Section getSection(String name) {
		Iterator i = sections.iterator();
		while (i.hasNext()) {
			Section next = (Section)i.next();
			if ( next.getName().equals(name) ) {
				return next;
			}
		}

		return null;
	}

	public void toString(String prefix, StringBuffer s) {
		s.append(prefix + "Machine:              " + getMachine() + "\n");
		s.append(prefix + "NumberOfSections:     " + NumberOfSections + "\n");
		s.append(prefix + "TimeDateStamp:        " + TimeDateStamp + "\n");
		s.append(prefix + "PointerToSymbolTable: " + PointerToSymbolTable + "\n");
		s.append(prefix + "NumberOfSymbols:      " + NumberOfSymbols + "\n");
		s.append(prefix + "SizeOfOptionalHeader: " + SizeOfOptionalHeader + "\n");
		s.append(prefix + "Characteristics:      " + CrystalUtils.toHex(Characteristics) + "\n");

		s.append(prefix + "String Table:\n");
		strings.toString(prefix + "   ", s);

		Iterator j = symbols.iterator();
		for (int x = 0; j.hasNext(); x++) {
			Symbol next = (Symbol)j.next();
			s.append(prefix + "Symbol " + x + "\n");
			next.toString(prefix + "   ", s);
		}

		Iterator i = sections.iterator();
		for (int x = 1; i.hasNext(); x++) {
			Section next = (Section)i.next();
			s.append(prefix + "Section " + x + "\n");
			next.toString(prefix + "   ", s);
		}
	}

	public String toString() {
		StringBuffer buffer = new StringBuffer();
		toString("", buffer);
		return buffer.toString();
	}
}
