package crystalpalace.coff;

import crystalpalace.util.*;
import java.util.*;

public class Section extends COFFPart {
	protected byte[] Name;
	protected long   VirtualSize;
	protected long   VirtualAddress;
	protected long   SizeOfRawData;
	protected long   PointerToRawData;
	protected long   PointerToRelocations;
	protected long   PointerToLinenumbers;
	protected int    NumberOfRelocations;
	protected int    NumberOfLinenumbers;
	protected SectionFlags Characteristics;

	/* the data for the section */
	protected byte[] RawData;

	/* our relocations */
	protected List   relocations = new LinkedList();

	public String getName() {
		ByteWalker walker = new ByteWalker(Name);

		String _name = walker.readStringA(8);
		if (_name.length() == 0)
			return "";

		if (_name.charAt(0) == '/') {
			String rest = _name.substring(1);
			int    pos  = CrystalUtils.parseInt(rest, -1);
			if (pos == -1)
				return _name;
			return getObject().getStringAt(pos);
		}

		return _name;
	}

	/* returns our data, aligned to a 4KB page size */
	public byte[] getData() {
		int    rawlen = RawData.length;
		int    padlen = 4096 - (rawlen % 4096);

		byte[] temp   = new byte[rawlen + padlen];
		for (int x = 0; x < RawData.length; x++) {
			temp[x] = RawData[x];
		}

		return temp;
	}

	/* we're going to clue off the characteristics value here, but really this is
	 * a check for the .bss section */
	public boolean isUninitialized() {
		return Characteristics.isUninitialized();
	}

	public int getPadLength() {
		return 4096 - (RawData.length % 4096);
	}

	/* returns our data without the aligning */
	public byte[] getRawData() {
		return RawData;
	}

	/* patch the specified byte array into our section raw data, thanks */
	public void patch(int offset, byte[] patch) {
		for (int x = 0; x < patch.length; x++) {
			RawData[x + offset] = patch[x];
		}
	}

	public long getVirtualSize() {
		return VirtualSize;
	}

	public long getVirtualAddress() {
		return VirtualAddress;
	}

	public List getRelocations() {
		return relocations;
	}

	public Section(COFFObject o, ByteWalker walker) {
		super(o);

		Name                 = walker.popBytes(8);
		VirtualSize          = walker.readInt();
		VirtualAddress       = walker.readInt();
		SizeOfRawData        = walker.readInt();
		PointerToRawData     = walker.readInt();
		PointerToRelocations = walker.readInt();
		PointerToLinenumbers = walker.readInt();
		NumberOfRelocations  = walker.readShort();
		NumberOfLinenumbers  = walker.readShort();
		Characteristics      = new SectionFlags(o, walker);

		/* if we're not init, we're probably the .bss section. */
		if (Characteristics.isUninitialized()) {
			RawData              = new byte[(int)SizeOfRawData];
		}
		/* read the raw data for the section */
		else {
			walker.GoTo((int)PointerToRawData);
			RawData              = walker.getBytes((int)SizeOfRawData);
			walker.Return();
		}

		/* read the relocations */
		walker.GoTo((int)PointerToRelocations);
		for (int x = 0; x < NumberOfRelocations; x++) {
			relocations.add(new Relocation(o, walker));
		}
		walker.Return();
	}

	public void toString(String prefix, StringBuffer s) {
		s.append(prefix + "Name:                 " + getName() + "\n");
		s.append(prefix + "VirtualSize:          " + VirtualSize + "\n");
		s.append(prefix + "VirtualAddress:       " + CrystalUtils.toHex(VirtualAddress) + "\n");
		s.append(prefix + "SizeOfRawData:        " + SizeOfRawData + "\n");
		s.append(prefix + "PointerToRawData:     " + PointerToRawData + "\n");
		s.append(prefix + "PointerToRelocations: " + PointerToRelocations + "\n");
		s.append(prefix + "PointerToLinenumbers: " + PointerToLinenumbers + "\n");
		s.append(prefix + "NumberOfRelocations:  " + NumberOfRelocations + "\n");
		s.append(prefix + "NumberOfLineNumbers:  " + NumberOfLinenumbers + "\n");
		s.append(prefix + "Characteristics:      " + Characteristics + "\n");

		Iterator i = relocations.iterator();
		for (int x = 0; i.hasNext(); x++) {
			Relocation next = (Relocation)i.next();
			s.append(prefix + "Relocation " + x + "\n");
			next.toString(prefix + "   ", s);
		}
	}
}
