package crystalpalace.export;

import crystalpalace.coff.*;
import crystalpalace.util.*;

import java.util.*;
import java.nio.*;

/*
 * What do we do here? We pack in values, resolve symbols, blah blah blah
 */
public abstract class ExportObject {
	protected COFFObject object;
	protected Concat     concat    = new Concat();
	protected Map        sections  = new LinkedHashMap();

	private class ExportSection {
		int        offset;
		int        length;
		String     name;
		ByteBuffer buffer;

		public ExportSection(String name, byte[] data, int padlen) {
			this.name   = name;
			this.offset = concat.length();
			this.buffer = ByteBuffer.wrap(data);
			this.length = data.length;

			concat.add(data);

			/* why do we do this? For an important reason. We want the section offsets to represent the
                	   final layout of these objects in memory. BUT, we  want the actual combined package to
			   not include this padding. This is to keep things... teensy tinsy. */
			if (padlen > 0) {
				concat.pad(padlen);
				length += padlen;
			}

			String machine = object.getMachine();
			if ("x64".equals(machine))
                        	buffer.order(ByteOrder.LITTLE_ENDIAN);
			else if ("x86".equals(machine))
                        	buffer.order(ByteOrder.LITTLE_ENDIAN);
		}

		public int getLength() {
			return length;
		}

		public int getOffset() {
			return offset;
		}

		public ByteBuffer getBuffer() {
			return buffer;
		}
	}

	public void patch(String symb, byte[] data) {
		object.patch(symb, data);
	}

	public ExportObject(COFFObject object) {
		this.object = object;

		addSections();
	}

	public abstract void addSections();

	protected boolean hasOffset(String name) {
		return sections.containsKey(name);
	}

	protected int getOffset(String name) {
		if (sections.get(name) == null)
			throw new RuntimeException("Section " + name + " does not exist in " + this.getClass().getSimpleName());
		return ((ExportSection)sections.get(name)).getOffset();
	}

	protected int getLength(String name) {
		return ((ExportSection)sections.get(name)).getLength();
	}

	protected int getLength() {
		return concat.length();
	}

	protected ByteBuffer getBuffer(String name) {
		return ((ExportSection)sections.get(name)).getBuffer();
	}

	public void addSectionRaw(String name) {
		if (object.getSection(name) != null)
			sections.put(name,  new ExportSection(name, object.getSection(name).getRawData(), 0) );
	}

	public void addSectionData(String name) {
		if (object.getSection(name) != null)
			sections.put(name,  new ExportSection(name, object.getSection(name).getRawData(), object.getSection(name).getPadLength()) );
	}

	public void addSectionEmpty(String name) {
		if (object.getSection(name) != null) {
			sections.put(name,  new ExportSection(name, new byte[0], object.getSection(name).getRawData().length));
			Logger.print_stat("Adding: " + name + " as empty section, saving: " + object.getSection(name).getRawData().length + "b");
		}
		/*
		 * We do this for good reason. We need the offset to .bss (what this is used for) to calculate a virtual and read length of our data
		 * for PICOs and BOFFs. So, when we add an empty section, we really need it to exist... to safely reference the length later.
		 */
		else {
			sections.put(name, new ExportSection(name, new byte[0], 0));
		}
	}

	/* this is public so we can add variables to our PIC pack that reference our rDLL
	   or whatever else and "link" it to the .text section */
	public void addSection(String name, byte[] data) {
		if (sections.containsKey(name))
			throw new RuntimeException("Section " + name + " already exists");

		sections.put(name, new ExportSection(name, data, 0));
	}

	/* Is this piece a ready-to-run position-independent code or... what? */
	public abstract boolean isPIC();

	protected abstract void processRelocation(String section, Relocation r);

	/*
	 * Export our (as close as possible) ready-to-execute PIC.
	 *
	 * This is done by packing .text, .rdata, and other sections AND then resolving
	 * offsets to each within .text.
	 *
	 * Sadly, this is only really effective for x64 right now. x86 doesn't do
	 * relative addressing for data and so this offline linking isn't possible. :(
	 */
	public byte[] export() {
		Iterator i = object.getSection(".text").getRelocations().iterator();
		while (i.hasNext()) {
			Relocation r = (Relocation)i.next();
			processRelocation(".text", r);
		}

		return concat.get();
	}
}
