package crystalpalace.export;

import crystalpalace.coff.*;
import crystalpalace.util.*;

import java.util.*;
import java.nio.*;

/*
 * What do we do here? We pack in values, resolve symbols, blah blah blah
 */
public class ExportPICO extends ExportObject {
	protected List patchups;
	protected Map  functions;
	protected Map  apis;
	protected List apitable;

	public ExportPICO(COFFObject object) {
		super(object);
	}

	public void addSections() {
		/* setup our instance vars we need... need to do this here and not constructor as
		   this function is called by our parent class constructor */
		patchups  = new LinkedList();
		functions = new HashMap();
		apis      = new HashMap();

		/* our default "API" aka functions we resolve without a MODULE$ decorator */
		setAPI("LoadLibraryA, GetProcAddress");

		/* Now, let's setup the expected sections for our PICOs/COFFs */
		addSectionData(".text");

		addFunctionPtrTable();

		addSectionRaw(".rdata");
		addSectionRaw(".data");

		/* we're going to add .bss RIGHT before we export. See
		   comments below on why. */
	}

	/*
	 * So, we're going to have a PICO concept of "APIs". An API is a function
	 * that is not decorated (e.g., not LIB$ attached to it). And, the PICO loader
	 * will receive this library as an extension of IMPORTFUNCS. We are going to treat
	 * that first parameter as basically an array of pointers, because that's EXACTLY
	 * what it is. ANYWAYS... if something is expected to be in that array of pointers
	 * we are going to... very nicely... y'know... index into that to handle the linking
	 * of that pointer into our GOT for function pointers. A nice clean way of setting
	 * up a contract for resolving internal APIs, making it expandable, and most importantly
	 * NOT PACKAGING strings that give away too easily what we're resolving.
	 */
	public int getAPI(String _symbol) {
		Iterator i = apitable.iterator();
		for (int x = 0; i.hasNext(); x++) {
			String sym = (String)i.next();
			if (sym.equals(_symbol))
				return x;
		}

		return -1;
	}

	/* set the API for this PICO */
	public void setAPI(String symbols) {
		/* initialize this */
		apitable = new LinkedList();

		/* add our values to it */
		String[] blah = symbols.split(",\\s*");
		for (int x = 0; x < blah.length; x++) {
			apitable.add(blah[x].trim());
		}

		/* and then do some sanity checks */
		if (getAPI("LoadLibraryA") != 0)
			throw new RuntimeException("LoadLibraryA is required as the first API entry.");

		if (getAPI("GetProcAddress") != 1)
			throw new RuntimeException("GetProcAddress is required as the second API entry.");
	}

	// https://devblogs.microsoft.com/oldnewthing/20060727-00/?p=30343
	public void addFunctionToResolve(String _symbol) {
		String symbol = _symbol;

		if (symbol.startsWith("__imp__")) {
			symbol = symbol.substring("__imp__".length());
		}
		else if (symbol.startsWith("__imp_")) {
			symbol = symbol.substring("__imp_".length());
		}

		/* extract our module and function from the name */
		String module;
		String func;

		String parse[] = symbol.split("\\$");
		if (parse.length == 2) {
			module = parse[0];
			func   = parse[1];
		}
		else {
			module = "";
			func   = symbol;
		}

		/* get rid of the @# stuff at the end */
		parse = func.split("\\@");
		if (parse.length == 2)
			func   = parse[0];

		/* If we don't have a module, then we're an (internal) API */
		if ("".equals(module)) {
			apis.put(func, _symbol);
		}
		/* If we do have a module, then we're a Win32 LIB$Func API */
		else {
			if (!functions.containsKey(module)) {
				functions.put(module, new HashMap());
			}

			Map table = (Map)functions.get(module);
			table.put(func, _symbol);
		}
	}

	protected void processFunctionRelocation(String symbol, Relocation r) {
		if (hasOffset(symbol)) {
			/* no need to re-add to our symbol table, it's already there */
		}
		else if (r.is("x64", Relocation.IMAGE_REL_AMD64_REL32)) {
			addSection(symbol, new byte[8]);
			addFunctionToResolve(symbol);
		}
		else if (r.is("x86", Relocation.IMAGE_REL_I386_DIR32)) {
			addSection(symbol, new byte[4]);
			addFunctionToResolve(symbol);
		}
	}

	public void addFunctionPtrTable() {
		Iterator i = object.getSection(".text").getRelocations().iterator();
		while (i.hasNext()) {
			Relocation r      = (Relocation)i.next();
			String     symbol = r.getSymbolName();
			if (symbol.startsWith("__imp__")) {
				processFunctionRelocation(symbol, r);
			}
			else if (symbol.startsWith("__imp_")) {
				processFunctionRelocation(symbol, r);
			}
		}
	}

	/*
	 * PICOs allow the .text and data sections to exist separate of eachother. Our code for tracking
	 * offsets and addresses assumes .text and data are contiguous. So, to make all of this work,
	 * we're subtracking the virtual size (size in memory, when loaded) of .text from our symbol
	 * offsets. The result is the offset of the data, relative to the base memory where our data lives.
	 */
	public int getDataOffset(String sym) {
		/* special case, esp. for x86, there's no section offset between .text and itself */
		if (".text".equals(sym))
			return 0;

		return getOffset(sym) - getLength(".text");
	}

	protected void processRelocation(String section, Relocation r) {
		if (!hasOffset(r.getSymbolName())) {
			throw new RuntimeException("Can't process relocation for " + r.toSimpleString());
		}
		else if (r.is("x64", Relocation.IMAGE_REL_AMD64_REL32)) {
			long reladdress = getDataOffset( r.getSymbolName() ) - ( r.getVirtualAddress() + 4 );
			int  offset     = getBuffer(section).getInt((int)r.getVirtualAddress());

			getBuffer(section).putInt((int)r.getVirtualAddress(), (int)reladdress + offset);

			patchups.add(r);
		}
		else if (r.is("x86", Relocation.IMAGE_REL_I386_DIR32)) {
			long offsetSymbolFromBase = getDataOffset( r.getSymbolName() );
			int  offsetDataFromSymbol = getBuffer(section).getInt((int)r.getVirtualAddress());

			Logger.println("Symbol " + r.getSymbolName() + " from base? " + offsetSymbolFromBase + ", data from symbol? " + offsetDataFromSymbol);

			getBuffer(section).putInt((int)r.getVirtualAddress(), (int)offsetSymbolFromBase + offsetDataFromSymbol);

			patchups.add(r);
		}
		else {
			throw new RuntimeException("Can't offline process relocation " + r);
		}
	}

	public byte[] export() {
		/*
		 * We're adding .bss at the end and making sure it's the last thing, why? As a space-saving measure...
		 * we send over a size for our final data that's different from the actual packed data we're sent. That
		 * difference in space is our zeroed out .bss section. .bss is unintialized global variables, for those
		 * following at home. If we don't do things this way, what happens is an end-user might link something
		 * from a .spec into this COFF and... if our .bss supposedly came earlier, then we'll have the zero length
		 * (in actuality) bss content get stomped over by the linked data. Putting it at the end makes sure our
		 * .bss offsets/references are always in that empty slack space at the end. That's what's going on here.
		 */
		addSectionEmpty(".bss");

		/*
		 * Now, on with the show, we're building up a packed structure of "directives", parsed by pico.h to
		 * simplify loading and resolving references within our PICO. Following this is the actual .text section
		 * And a packed and padded(? did I make that work) combination of our various linked sections after that.
		 */
		byte[] content = super.export();
		byte[] program = getLoaderProgram();
		byte[] header  = getHeader(program, content);

		Concat c = new Concat();
		c.add(header);
		c.add(program);
		c.add(content);

		Logger.print_stat("export() header " + header.length + "b, program " + program.length + "b, content " + content.length + "b");

		return c.get();
        }

	public String getEntrySymbol() {
		String machine = object.getMachine();
		if ("x64".equals(machine))
			return "go";
		else if ("x86".equals(machine))
			return "_go";

		throw new RuntimeException(machine + " machine is not supported");
	}

	/*
	 * Generate our header content, compatbile with our PICO runner code.
	 *
	 * typedef struct {
	 * 	int codeLength;
	 * 	int dataLength;
	 * 	int rsrcOffset;
	 *	int entryAddress;
	 * } PICO_HDR;
	 *
 	 */
	public byte[] getHeader(byte[] program, byte[] content) {
		Packer temp = new Packer();
		temp.little();

		temp.addInt(getLength(".text"));					/* codeLength */
		temp.addInt(getLength() - getLength(".text"));				/* dataLength */
		temp.addInt(program.length + 16);					/* rsrcOffset */
		temp.addInt((int)object.getSymbol(getEntrySymbol()).getValue());	/* entryAddress */

		int enter = (int)object.getSymbol(getEntrySymbol()).getValue();

		Logger.println("PICO Header");
		Logger.println("Length: " + getLength());
		Logger.println("  Code: " + getLength(".text"));
		Logger.println("  Data: " + (getLength() - getLength(".text")));
		Logger.println("Offset: " + (program.length + 12));
		Logger.println("Entry:  " + enter + " " + CrystalUtils.toHex(enter));

		return temp.getBytes();
	}

	/* these are the directives that tell the loader how to load our PICO */
	public byte[] getLoaderProgram() {
		ProgramPICO program = new ProgramPICO();

		/* add our copy instruction to the program, first and foremost */
		int codeRealLen  = object.getSection(".text").getRawData().length;
		int codeVirtLen  = getLength(".text");
		int dataVirtLen  = getLength() - codeVirtLen;
		int dataRealLen  = dataVirtLen - getLength(".bss");

			/* copy our .text section into the first pages */
		program.CopyCode(0, 0, codeRealLen);

			/* copy all of our data into the following poages */
		program.CopyData(codeRealLen, 0, dataRealLen);

		Logger.println("CODE is at: 0 to " + codeRealLen);
		Logger.println("DATA is at: " + codeVirtLen + " (real: " + codeRealLen + ") + " + dataVirtLen + " (real: " + dataRealLen + ")");

		/* put in our base patchups */
		Iterator i = patchups.iterator();
		while (i.hasNext()) {
			Relocation r = (Relocation)i.next();

			if (r.is("x64", Relocation.IMAGE_REL_AMD64_REL32)) {
				program.PatchBaseDiff((int)r.getVirtualAddress());
			}
			else if (r.is("x86", Relocation.IMAGE_REL_I386_DIR32)) {
				if ( ".text".equals(r.getSymbolName()) ) {
					/* we need to patch in the base address of our .text section, for function pointers and such */
					program.PatchText((int)r.getVirtualAddress());
				}
				else {
					program.PatchBase((int)r.getVirtualAddress());
				}
			}
		}

		/* next, handle our import table */
		Iterator j = functions.entrySet().iterator();
		while (j.hasNext()) {
			Map.Entry entry  = (Map.Entry)j.next();
			String    module = (String)entry.getKey();
			Map       funcs  = (Map)entry.getValue();

			program.LoadLibrary(module);

			Iterator k = funcs.entrySet().iterator();
			while (k.hasNext()) {
				Map.Entry fentry = (Map.Entry)k.next();
				String    func   = (String)fentry.getKey();
				String    sym    = (String)fentry.getValue();

				program.GetProcAddress(func);
				program.PatchFunction(getDataOffset(sym));
			}
		}

		/* let's handle our (internal) API calls */
		Iterator m = apis.entrySet().iterator();
		while (m.hasNext()) {
			Map.Entry entry  = (Map.Entry)m.next();
			String    func   = (String)entry.getKey();
			String    sym    = (String)entry.getValue();
			int       index  = getAPI(func);

			if (index == -1)
				throw new RuntimeException("Function " + func + " is not imported and not in MODULE$function format");

			program.PatchImport(getDataOffset(sym), index);
		}

		/* ... */
		program.Complete();

		return program.getBytes();
	}

	public boolean isPIC() {
		return false;
	}

	public static void main(String args[]) {
		if (args.length != 2) {
			CrystalUtils.print_error("./picobuild [/path/to/file.o] [out.bin]");
			return;
		}

		try {
			ExportPICO p = new COFFParser(CrystalUtils.readFromFile(args[0])).parse().exportObject();
			CrystalUtils.writeToFile(args[1], p.export());
		}
		catch (Exception ex) {
			CrystalUtils.handleException(ex);
		}
	}
}
