package crystalpalace.spec;

import crystalpalace.coff.*;
import crystalpalace.pe.*;
import crystalpalace.util.*;
import crystalpalace.export.*;

import java.util.*;
import java.io.*;

public class SpecProgram {
	protected Map        directives = new HashMap();
	protected Stack      state      = new Stack();

	protected String     parent     = ".";
	protected String     name       = "";

	/* state information to aid error reporting */
	protected String     last       = "";
	protected SpecObject larg       = null;
	protected String     ltarg      = null;

	/* reset the state of this program, in case its used a second time */
	public void reset() {
		state = new Stack();
		last  = "";
		larg  = null;
		ltarg = null;
	}

	public SpecProgram(String parent) {
		this.parent = parent;

		/* make the default name of the spec, the filename only */
		name = new java.io.File(parent).getName();
	}

	public String getFile() {
		return parent;
	}

	public String getLastTarget() {
		return ltarg;
	}

	public SpecObject getLastArgument() {
		return larg;
	}

	public String getLastCommand() {
		return last;
	}

	public Stack getStack() {
		return state;
	}

	public boolean targets(String arch) {
		return getInstructionsForLabel(arch).size() > 0;
	}

	public List getInstructionsForLabel(String label) {
		if (!directives.containsKey(label))
			directives.put(label, new LinkedList());

		return (LinkedList)directives.get(label);
	}

	protected void push(byte[] b, String source) {
		state.push( new SpecObject(this, b, source) );
		larg = null;
	}

	protected void push(ExportObject o, String source) {
		state.push( new SpecObject(this, o, source) );
		larg = null;
	}

	protected void push(SpecObject obj) {
		state.push(obj);
		larg = null;
	}

	protected SpecObject pop() throws SpecProgramException {
		if (state.empty())
			throw new SpecProgramException(this, "POP - stack is empty");

		larg = (SpecObject)state.pop();

		return larg;
	}

	public byte[] run(Map env) throws SpecProgramException, SpecParseException {
		byte[]   dll = getFromEnv(env, "$DLL");
		PEObject obj = new PEParser(dll).parse();

		if (!obj.isValid())
			throw new RuntimeException("Invalid DLL: " + obj.getLastError());

		String label = obj.getMachine();

		if (!targets(label))
			throw new RuntimeException("Spec does not have target for '" + label + "'");

		return run(label, env);
	}

	public byte[] run(String label, Map env) throws SpecProgramException, SpecParseException {
		List  inst  = getInstructionsForLabel(label);

		/* set this, please */
		ltarg = label;

		if (inst.size() == 0)
			throw new RuntimeException("No spec for label '" + label + "'");

		Iterator i = inst.iterator();
		while (i.hasNext()) {
			CommandParser command = (CommandParser)i.next();
			String        args[]  = command.getArguments();

			last = command.getOriginal();

			// load [file]          BYTES    -		Read file and push content onto stack
			if ("load".equals(command.getCommand()) && args.length == 1) {
				try {
					File temp = getFileFromArg(args[0]);
					push( CrystalUtils.readFromFile( temp.getPath() ), temp.getName() );
				}
				catch (IOException ioex) {
					throw new SpecProgramException(this, ioex.getMessage());
				}
			}
			else if ("load".equals(command.getCommand()) && args.length == 2) {
				try {
					File temp = getFileFromArg(args[1]);
					putEnv(env, args[0], CrystalUtils.readFromFile(temp.getPath()));
				}
				catch (IOException ioex) {
					throw new SpecProgramException(this, ioex.getMessage());
				}
			}
			else if ("run".equals(command.getCommand())) {
				try {
					File temp = getFileFromArg(args[0]);

					/* create a NEW linkspec */
					LinkSpec spec    = LinkSpec.Parse(temp.getPath());

					/* run the .spec */
					push( spec.program.run(label, env), temp.getName() );
				}
				catch (IOException ioex) {
					throw new SpecProgramException(this, ioex.getMessage());
				}
			}
			// link "section"       OBJECT  OBJECT, BYTES	Link BYTES to section in OBJECT
			else if ("link".equals(command.getCommand())) {
				byte[]       value = pop().getBytes();
				SpecObject   obj   = pop();

				try {
					obj.getObject().addSection(args[0], value);
				}
				catch (RuntimeException rex) {
					handleException(rex);
				}

				push(obj);
			}
			// export               -	OBJECT     	export object, return
			else if ("export".equals(command.getCommand())) {
				SpecObject   obj  = pop();
				ExportObject eobj = obj.getObject();
				try {
					push( eobj.export(), obj.getSource() );
				}
				catch (RuntimeException rex) {
					handleException(rex);
				}
			}
			// make object		OBJECT  BYTES		Turn FILE into ready-to-export PICO object
			// make pic 	        OBJECT  BYTES		Turn FILE into ready-to-export PIC object
			else if ("make pic".equals(command.getOriginal())) {
				SpecObject obj = pop();
				COFFParser parser = createParser(obj);
				push( parser.export(), obj.getSource() );
			}
			else if ("make pic64".equals(command.getOriginal())) {
				if (!"x64".equals(label))
					throw new SpecProgramException(this, "make pic64 is x64-only");

				SpecObject obj = pop();
				COFFParser parser = createParser(obj);
				push( parser.export64(), obj.getSource() );
			}
			else if ("make object".equals(command.getOriginal())) {
				SpecObject obj = pop();
				COFFParser parser = createParser(obj);
				push( parser.exportObject(), obj.getSource() );
			}
			/* generate a random string and stick it into an environment var */
			else if ("generate".equals(command.getCommand())) {
				String var  = args[0];
				int    len  = CrystalUtils.parseInt(args[1], 0);
				if (len < 0)
					throw new SpecProgramException(this, "Nice try, we can't generate a negative size byte array");

				byte[] val  = new byte[len];

				try {
					java.security.SecureRandom rng = java.security.SecureRandom.getInstanceStrong();
					rng.nextBytes(val);
				}
				catch (Exception ex) {
					throw new SpecProgramException(this, ex.getMessage());
				}

				putEnv(env, var, val);
			}
			/* patch a symbol inside of one of our COFF's sections */
			else if ("patch".equals(command.getCommand())) {
				String var = args[1];
				String sym = args[0];

				SpecObject obj  = pop();

				if (!"x64".equals(label) && obj.getObject().isPIC())
					throw new SpecProgramException(this, "Can't patch symbols in non-x64 PIC");

				try {
					obj.getObject().patch(sym, getFromEnv(env, var));
				}
				catch (RuntimeException rex) {
					handleException(rex);
				}

				push(obj);
			}
			else if ("import".equals(command.getCommand())) {
				String     funcs = args[0];
				SpecObject obj   = pop();

				if (obj.getObject() instanceof ExportPICO) {
					try {
						((ExportPICO)obj.getObject()).setAPI(funcs);
					}
					catch (RuntimeException rex) {
						handleException(rex);
					}
				}
				else {
					throw new SpecProgramException(this, "Argument is not a COFF - can't import functions to it");
				}

				push(obj);
			}
			else if ("xor".equals(command.getCommand())) {
				SpecObject obj = pop();

				byte[] key = getFromEnv(env, args[0]);
				byte[] val = obj.getBytes();

				for (int x = 0; x < val.length; x++) {
					val[x] ^= key[x % key.length];
				}

				push( val, obj.getSource() );
			}
			else if ("rc4".equals(command.getCommand())) {
				SpecObject obj = pop();

				byte[] key = getFromEnv(env, args[0]);
				byte[] val = obj.getBytes();

				try {
					val = CrystalUtils.rc4encrypt(key, val);
				}
				catch (Exception ex) {
					throw new SpecProgramException(this, ex.getMessage());
				}

				push( val, obj.getSource() );
			}
			// prepend [length] to our [data] on the stack
			else if ("preplen".equals(command.getCommand())) {
				SpecObject obj = pop();

				byte[] data = obj.getBytes();

				Packer pack = new Packer();
				pack.little();
				pack.addData((byte[])data);	/* prepends our length for us */

				push(pack.getBytes(), obj.getSource());
			}
			else if ("prepsum".equals(command.getCommand())) {
				SpecObject obj = pop();

				byte[] data = obj.getBytes();

				Packer pack = new Packer();
				pack.little();
				pack.addDataVerify((byte[])data); /* prepends adler32sum */

				push(pack.getBytes(), obj.getSource());
			}
			// push $VAR            BYTES   -               Push $VAR content onto stack
			else if ("push".equals(command.getCommand())) {
				push(getFromEnv(env, args[0]), args[0]);
			}
			else {
				throw new SpecProgramException(this, "This is a bug? Did not process '" + command.getOriginal() + "'");
			}
		}

		if (state.empty())
			throw new SpecProgramException(this, "Stack is empty. Where's the bytes I want to return?");

		SpecObject obj = pop();

		if (!state.empty())
			throw new SpecProgramException(this, "Stack is not empty. Make sure all objects are processed");

		return obj.getBytes();
	}

	/* Since I was too lazy to create a COFF[Verb]Exception class; I wrap all of my various COFF-related errors as
	 * unchecked RuntimeExceptions. Works great, except when I have a bonafide... oops.... I didn't mean to throw that
	 * type issue. So, we're checking here if the exception is a RuntimeException or something else that needs more
	 * context to aid fixing it. */
	public void handleException(RuntimeException rex) throws SpecProgramException {
		if (rex.getClass() != RuntimeException.class) {
			CrystalUtils.handleException(rex);
			throw new SpecProgramException(this, rex.getClass().getName() + ": " + rex.getMessage());
		}
		else {
			throw new SpecProgramException(this, rex.getMessage());
		}
	}

	public COFFParser createParser(SpecObject obj) throws SpecProgramException {
		COFFParser parser = new COFFParser(obj.getBytes());
		try {
			parser.parse();
		}
		catch (RuntimeException rex) {
			handleException(rex);
		}

		if (! ltarg.equals(parser.getObject().getMachine()) )
			throw new SpecProgramException(this, parser.getObject().getMachine() + " COFF arch differs from " + ltarg + " .spec target");

		return parser;
	}

	public File getFileFromArg(String arg) throws SpecProgramException {
		File temp = new File( new File(parent).getParent(), arg );
		if (!temp.exists())
			throw new SpecProgramException(this, "File does not exist " + temp.getPath());

		if (temp.isDirectory())
			throw new SpecProgramException(this, "File is a folder " + temp.getPath());

		if (!temp.canRead())
			throw new SpecProgramException(this, "File is not readable " + temp.getPath());

		return temp;
	}

	public void putEnv(Map env, String key, byte[] data) throws SpecProgramException {
		if (env.containsKey(key))
			throw new SpecProgramException(this, key + " is already present in environment. Can't overwrite");
		else
			env.put(key, data);
	}

	public byte[] getFromEnv(Map env, String key) throws SpecProgramException {
		if (env.containsKey(key)) {
			Object temp = env.get(key);

			if (temp.getClass().isArray() && temp.getClass().getComponentType() == byte.class) {
				return (byte[])temp;
			}
			else {
				throw new SpecProgramException(this, "Var " + key + " is not a byte[]");
			}
		}
		else {
			throw new SpecProgramException(this, "Var " + key + " is not set");
		}
	}
}
